#
#
# 1... BIVPOIS.TABLE :	Bivariate Poisson Probability Function (in Tabular form) using 
#			retrospective relationships
# 2... BIVPOIS	     :  Probability function of Bivariate Poisson
# 3... LOGBP	     :  Logarithm of the Probability function of Bivariate Poisson
# 4... SIMPLE.BP     :  EM for fitting a simple bivariate poisson model with no covariates 
#			on lambda1-3
# 5... GLM.BP	     :	EM for fitting a GL bivariate poisson model with covariates 
# 6... GLM.DIBP	     :	EM for fitting a Diagonal Inflated bivariate poisson model 
#		        with covariates 
# ---------------------------------------------------------------------------------------------




# ---------------------------------------------------------
# 1... Bivariate poisson table probability function
# ---------------------------------------------------------

bivpois.table<-function(x, y, lambda = c(1, 1, 1))
{
# x 1st count variable
# y 2nd count variable
#------------------------------
	j_0
	n <- length(x)
	maxy <- c(max(x), max(y))	#Set initial values for parameters
	lambda1 <- lambda[1]
	lambda2 <- lambda[2]
	lambda3 <- lambda[3]
	if((x == 0) | (y == 0)) {
		prob <- matrix(NA, nrow = maxy[1] + 1, ncol = maxy[2]+1, byrow = T)
		prob[maxy[1] + 1, maxy[2] + 1] <- exp( - lambda3) * 
			dpois(x[j], lambda1[j]) * dpois(y[j], lambda2[j])
	}
	else {
		prob <- matrix(NA, nrow = maxy[1] + 1, ncol = maxy[2]+1, byrow = T)
		k <- 1
		m <- 1
		prob[k, m] <- exp( - lambda1 - lambda2 - lambda3)
		for(i in 2:(maxy[1] + 1)) {
			prob[i, 1] <- (prob[i - 1, 1] * lambda1)/(i - 1)
		}
		for(j in 2:(maxy[2] + 1)) {
			prob[1, j] <- (prob[1, j - 1] * lambda2)/(j - 1)
		}
		for(j in 2:(maxy[2] + 1)) {
			for(i in 2:(maxy[1] + 1)) {
				prob[i, j] <- (lambda1 * prob[i - 1, j] + 
				  lambda3 * prob[i - 1, j - 1])/(i - 1)
			}
		}
	}
	result <- prob
	result
}
#end of function bivpois.table

# ---------------------------------------------------------
# 2... Bivariate poisson probability function
# ---------------------------------------------------------


bivpois<-function(x, y, lambda = c(1, 1, 1)) {
	lbp<-logbp(x, y, lambda = c(1, 1, 1))
	result<-exp(lbp)
	result
}



# ---------------------------------------------------------
# 3... Bivariate poisson log-probability function
# ---------------------------------------------------------



logbp<-function(x, y, lambda = c(1, 1, 1))
{
# x 1st count variable
# y 2nd count variable
#------------------------------
	n <- length(x)
#
	x0_x[1]
	y0_y[1]
	xymin<-min( x0,y0 )
	lambdaratio<-lambda[3]/(lambda[1]*lambda[2])
#	
	i_0:xymin
	sums_ -lgamma(x-i+1)-lgamma(i+1)-lgamma(y-i+1)+i*log(lambdaratio)
	maxsums <- max(sums)
	sums<- sums - maxsums
	logsummation<- log( sum(exp(sums)) ) + maxsums 
	logbp<- -sum(lambda) + x * log( lambda[1] ) + y * log( lambda[2] ) + logsummation 
	logbp
}
 

#end of function bivpois



# ------------------------------------------------------------------------
# 4... EM for fitting a simple bivariate poisson model with no covariates on lambda1-3
# ------------------------------------------------------------------------

simple.bp<-function(x, y, nmax=300, pres=0.00000001, ini3=1.0)
{
#
# Karlis and Ntzoufras (2003)
# EM algorithms for Bivariate Poisson Models
#
# x: 1st vector of counts
# y: 2nd vector of counts
# nmax = maximum number of iterations 
#
# Data length
n<-length(x)
#
#
#
# initial values
s<-rep(0,n)
like<-1:n*0
zero<- ( x==0 )|( y==0 )
#
#
#
# Initial values for lambda

lambda3<- ini3
lambda1<- max( 0.1, mean(x)-lambda3 )
lambda2<- max( 0.1, mean(y)-lambda3 )
#
#
difllike<-1000.0
loglike0<-1000.0
i<-0
loglike<-rep(0,nmax)
while ( (difllike>pres) & (i <= nmax) ) {
	i<-i+1
	#####   E step  ######
	for (j in 1:n) {
		if (zero[j]) {
			s[j]<-0;
			like[j]<- log(dpois(x[j], lambda1)) + log(dpois(y[j], lambda2))-lambda3;
			}
		else {
			lbp1<- logbp( x[j]-1, y[j]-1, lambda=c(lambda1,lambda2,lambda3) );
			lbp2<- logbp( x[j]  , y[j]  , lambda=c(lambda1,lambda2,lambda3) );
			
			s[j]<-exp( log(lambda3) + lbp1 - lbp2 );
			like[j]<-lbp2;
		}
	}
	##### end of E step  ######
	x1<-x-s
	x2<-y-s
	loglike[i]<-sum(like)
	difllike<-abs( (loglike0-loglike[i])/loglike0 )
	loglike0<-loglike[i]
	#
	#
	#####   M step  ######
	#
	# 	fit model on lambda3
	lambda1<-mean(x1)
	lambda2<-mean(x2)
	lambda3<-mean(s)
	#####   end of M step  ######
	printpars<-c(i,lambda1, lambda2, lambda3, ,loglike[i] )
	names(printpars)<-c('Iter.', 'lambda1', 'lambda2', 'lambda3','loglike' )
	print( round(printpars ,2 ) )
	print( paste('Relative Difference in Loglike:', difllike) ) 
	}
#
#	calculation of BIC and AIC of Bivariate Poisson model
	noparams<- 3 
	AIC<- -2*loglike[i] + noparams * 2
	BIC<- -2*loglike[i] + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-0.000000000001
	y.mean<-y
	y.mean[y==0]<-0.000000000001
	AIC.sat <-  sum( log( dpois( x , x.mean ) ) + log( dpois(y , y.mean) ) )
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#		
#	Calculation of BIC, AIC of simple Poisson model
	x.mean<-mean(x)
	y.mean<-mean(y)
	AIC.pois <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.pois <-  -2 * AIC.pois + 2* log(2*n)
	AIC.pois <-  -2 * AIC.pois + 2* 2
	
	AICtotal<-c(AIC.sat, AIC.pois, AIC)
	BICtotal<-c(BIC.sat, BIC.pois, BIC )
	
	names(AICtotal)<- c( 'Saturated', 'Double Poisson', 'Bivar.Poisson' )
	names(BICtotal)<- c( 'Saturated', 'Double Poisson', 'Bivar.Poisson' )
#
# Calculation of fitted values
	result<-list(lambda1=lambda1, lambda2=lambda2, lambda3=lambda3,loglikelihood=loglike[1:i],
			parameters=noparams, AIC=AICtotal, BIC=BICtotal ,iterations=i )
#
result
#
#
}







# ------------------------------------------------------------------------
# 5... EM for fitting a bivariate poisson model with covariates on lambda3
# ------------------------------------------------------------------------

glm.bp<-function(x, y, w1, w2, w3=0, l3=1, nmax=300, constant=c(1,1), pres=0.00000001)
{
#
# Karlis and Ntzoufras (2003)
# EM algorithms for Bivariate Poisson Models
#
# x: 1st vector of counts
# y: 2nd vector of counts
# w1, w2, w3 : data.frames with covariate of lambda1, lambda2, lambda3
#	Default value for w3 is constant.
#
#       w1,w2    should have same dimensions ( n x p = (sample size) x (# of covariates).
#		 if common parameters for lambda1 and lambda3 are needed for one
#		 variable then they should be in the same collumn of w1 and w2
#		 otherwize in different collumns 
#		 Example1 if  w1[,1]=sex w2[,1]=0 w1[,2]=0 w2[,2]=sex
#			  this will estimate different sex effect on lambda1 and lambda2
#		 alternative  w1[,1]=sex w2[,1]=sex w1[,2]=sex w2[,2]=0
#			      will estimate different sex effect and gives p.value for the
#				significance of their difference
#		Common Sex effect then: w1[,1]=sex and w2[,1]=sex 	
#
#          l3=0, assumes that l3=0           (fits double poisson model) ignoring w3
#          l3=1, assumes that l3 is constant (no covariates on lambda3) ignoring w3
#          l3=2, fits covariates (given by w3) on lambda3 
# nmax = maximum number of iterations 
# constant=c(1,1):Include Constant terms for lambda1-2 [1st value] and lambda3 [2nd value]
# 		Default values are positive for both 
#
#
# Data length
n<-length(x)
lengthpvec<-1
#
#
#
# initial values
s<-rep(0,n)
like<-1:n*0
zero<- ( x==0 )|( y==0 )
if   (l3==0) { lambda3<-rep(0,n) } 
else { lambda3<-rep( max(0.1, cov(x,y)), n) } 
#
#
p<-length(as.data.frame(w1))+1
p2<-length(as.data.frame(w3))+1
#
# form dataframes used
# data1 includes modelling on lambda1 and lambda2
# data2 includes modelling on lambda3 
#
data1<-data.frame( y0=c(x,y) )
data2<-data.frame( s = rep(0, n ) )
#
data1[,2:p]<-rbind( w1, w2 ) 
names(data1)[2:p]<-names(w1)
for (i in 2:p) {
	if ( is.factor(data1[,i])==T ){
		contrasts(data1[,i])<-contr.sum( length(levels(data1[,i])) )
	}
}
#
if (l3==1) {
	data2$con_rep(1,n) }
else if (l3==2) {
	data2[,2:p2]<-w3
	names(data2)[2:p2]<-names(w3)
	for (i in 2:p2) {
		if ( is.factor(data2[,i])==T ){
			contrasts(data2[,i])<-contr.sum(length(levels(data2[,i])))
		}
	}	

}

#
# Set formulas
if (constant[1]==1) { 
	data1$dcon<- c( rep(1,n),rep(0,n) )
	formula1<-'y0~.' }
else { formula1<-'y0~.-1' }
#
if      (l3==0) { formula0 <- 's~1'}
else if (l3==1) { formula0 <- 's~1'}
else            { 
	if (constant[2]==1) {formula0 <- 's~.'}
	else {formula0 <- 's~.-1'}
}
#
# Initial values for lambda

lambda<-glm(formula1,family=poisson, data=data1)$fitted
#
lambda1<-lambda[1:n]
lambda2<-lambda[(n+1):(2*n)]
#
difllike<-100.0
loglike0<-1000.0
i<-0

if (l3==0) {
	#
	# fit double poisson model
	y0<-c(x,y)
	m<-glm( formula1, family=poisson, data=data1 )
	p3<-length(m$coef)
	if (constant[1]==1) {
		beta<-c( m$coef[1]+m$coef[p3], m$coef[1:(p3-1)])
		names(beta)[1]<-'Intercept_1'
		names(beta)[2]<-'Intercept_2'	
	}
	else { beta<-m$coef }
	lambda<-fitted(m)
	lambda1<-lambda[1:n]
	lambda2<-lambda[(n+1):(2*n)]
	like<-dpois(x, lambda1) * dpois( y, lambda2 )
	loglike<-sum(log(like))
#
#	calculation of BIC and AIC for bivpoisson model
	noparams<- m$rank 
	AIC<- -2*loglike + noparams * 2
	BIC<- -2*loglike + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-0.000000000001
	y.mean<-y
	y.mean[y==0]<-0.000000000001
	AIC.sat <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#
	AICtotal<-c(AIC.sat, AIC); 
	BICtotal<-c(BIC.sat, BIC );
	names(AICtotal)<-c('Saturated', 'BP model')
	names(BICtotal)<-c('Saturated', 'BP model')

	result<-list(beta=beta, lambda3=0,lambda1=m$fitted[1:n], 	
		     lambda2=m$fitted[(n+1):(2*n)],loglikelihood=loglike, 
		     iterations=1, parameters=noparams,
		     AIC=AICtotal, BIC=BICtotal)
}
else {	
	loglike<-rep(0,nmax)
	while ( (difllike>pres) & (i <= nmax) ) {
	i<-i+1
	#####   E step  ######
	for (j in 1:n) {
		if (zero[j]) {
			s[j]<-0;
			like[j]<- log(dpois(x[j], lambda1[j]))+log(dpois(y[j],lambda2[j])) 				  -lambda3[j];
			}
		else {
			lbp1<-logbp(x[j]-1, y[j]-1, lambda=c(lambda1[j],lambda2[j],lambda3[j]) );
			lbp2<-logbp(x[j]  , y[j]  , lambda=c(lambda1[j],lambda2[j],lambda3[j]) );
#
			s[j]<- exp( log(lambda3[j]) + lbp1 - lbp2 );
			like[j]<-lbp2;
		}
	}
	##### end of E step  ######
	x1<-x-s
	x2<-y-s
	loglike[i]<-sum(like)
	difllike<-abs( (loglike0-loglike[i])/loglike0 )
	loglike0<-loglike[i]
	#
	#
	#####   M step  ######
	#
	# 	fit model on lambda3
	data2$s<-s
	m0<-glm( formula0, family=poisson, data=data2 )
	beta3<-m0$coef
	lambda3<-m0$fitted
	#
	#	fit model on lambda1 & lambda2
	data1$y0<-c(x1,x2)
	m<-glm( formula1, family=poisson, data=data1 )
	p3<-length(m$coef)
	if (constant[1]==1) {
		beta<-c( m$coef[1]+m$coef[p3], m$coef[1:(p3-1)])
		names(beta)[1]<-'Intercept_1'
		names(beta)[2]<-'Intercept_2'	
	}
	else { beta<-m$coef }
#	
#	
	lambda<-fitted(m)
	lambda1<-lambda[1:n]
	lambda2<-lambda[(n+1):(2*n)]
	#####   end of M step  ######
	printvector<-c( i, beta, beta3,loglike[i], difllike )
	names(printvector)<-c( 'iter', names(beta), paste('l3_',names(beta3),sep=''), 'loglike', 
				'Rel.Dif.loglike')
	lengthpvec<-length(printvector)
#	
	print( c( round( printvector[1:(lengthpvec-1)] ,2 ), printvector[lengthpvec] ) )
	}
#
#	calculation of BIC and AIC for bivpoisson model
	noparams<- m$rank + m0$rank
	AIC<- -2*loglike[i] + noparams * 2
	BIC<- -2*loglike[i] + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-0.000000000001
	y.mean<-y
	y.mean[y==0]<-0.000000000001
	AIC.sat <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#
	AICtotal<-c(AIC.sat, AIC); 
	BICtotal<-c(BIC.sat, BIC );
	names(AICtotal)<-c('Saturated', 'BP model')
	names(BICtotal)<-c('Saturated', 'BP model')

#
# Calculation of fitted values
	if (l3==1) {lambda3<-lambda3[1]}
	result<-list(beta=beta, beta3=beta3, lambda3=lambda3, lambda1=m$fitted[1:n], 
				lambda2=m$fitted[(n+1):(2*n)],loglikelihood=loglike[1:i],
				parameters=noparams, AIC=AICtotal, BIC=BICtotal,iterations=i )
#
#
} # end of elseif
#
result
#
#
}



#
#
# -------------------------------------------------------------------------------- 
# --------------------------------------------------------------------------------
# 6... EM for fitting a diagonal inflated bivariate poisson model with covariates 
# --------------------------------------------------------------------------------
# -------------------------------------------------------------------------------- 
#
glm.dibp<-function(x, y, w1, w2, w3=0, l3=1, nmax=300, constant=c(1,1), pres=0.00000001, distribution=1, jmax=2)
{
#
# Karlis and Ntzoufras (2003)
# EM algorithms for Diagonal inflated Bivariate Poisson Models
#
# x: 1st vector of counts
# y: 2nd vector of counts
# w1, w2, w3 : data.frames with covariate of lambda1, lambda2, lambda3
#	Default value for w3 is constant.
#
#       w1,w2    should have same dimensions ( n x p = (sample size) x (# of covariates).
#		 if common parameters for lambda1 and lambda3 are needed for one
#		 variable then they should be in the same collumn of w1 and w2
#		 otherwize in different collumns 
#		 Example1 if  w1[,1]=sex w2[,1]=0 w1[,2]=0 w2[,2]=sex
#			  this will estimate different sex effect on lambda1 and lambda2
#		 alternative  w1[,1]=sex w2[,1]=sex w1[,2]=sex w2[,2]=0
#			      will estimate different sex effect and gives p.value for the
#				significance of their difference
#		Common Sex effect then: w1[,1]=sex and w2[,1]=sex 	
#
#          l3=0, assumes that l3=0           (fits double poisson model) ignores w3
#          l3=1, assumes that l3 is constant (on covariates on lambda3) ignores w3
#          l3=2, fits covariates of w3 on lambda3 
# nmax = maximum number of iterations 
# Include Constant terms for lambda1-2 [1st value] and lambda3 [2nd value]
# 		Default values are positive for both 
#
#	distribution : Selection of diagonal inflation distribution
#		1: Discrete, jmax is the number of diagonal elements [0,1,...,]
#               2: Poisson 
#		3: Geometric
#	        Default is DISCRETE(2) with theta_0, theta_1, theta_2
# Data length
n<-length(x)
lengthprintvec<-1
#
#
#
# definition of diagonal inflated distribution
	maxy<-max(c(x,y))
	if ( distribution==1 ) { 
			dilabel<-paste('Inflation Distribution: Discrete with J=',jmax)
			if (jmax==0) {theta<-0}
			else { theta<-1:jmax*0+1/(jmax+1) }
			di.f_function (x, theta){
							JMAX<-length(theta)
							if      (x>JMAX) { res<-0 }
							else if (x==0)   { res<-1-sum(theta) }
							else             { res<-theta[x] } 
							res
					 }
	}
	else if ( distribution==2 ) { 
			dilabel<-'Inflation Distribution: Poisson'
			theta<-1.0;
			di.f_function (x, theta){
							if (theta>0) { res_dpois( x, theta ) }
							else {
								if (x==0) { res_1}
								else {res_0.000000000001}
							}
					 }
	}								 
	else {
			dilabel<-'Inflation Distribution: Geometric'
			theta<-0.5;
			di.f_function (x, theta){
							if (theta>0) { 
								if(theta==1) {theta<-0.9999999}
								res_dgeom( x, theta ) }
							else if (theta==1){
								if (x==0) { res_1}
								else {res_0.0000000001}
							}
							else {res_0.000000000001}
					 }
	}
# initial values
prob<-0.20
s<-rep(0,n)
vi_1:n*0
v1<-1-c(vi,vi)
like<-1:n*0
zero<- ( x==0 )|( y==0 )
if   (l3==0) { lambda3<-rep(0,n) } 
else { lambda3<-rep( max(0.1, cov(x,y)), n) } 
#
#
p<-length(as.data.frame(w1))+1
p2<-length(as.data.frame(w3))+1
#
# form dataframes used
# data1 includes modelling on lambda1 and lambda2
# data2 includes modelling on lambda3 
#
data1<-data.frame( y0=c(x,y) )
data2<-data.frame( s = rep(0, n ) )
#
data1[,2:p]<- rbind(w1,w2)
names(data1)[2:p]<-names(w1)
for (i in 2:p) {
	if ( is.factor(data1[,i])==T ){
		contrasts(data1[,i])<-contr.sum( length(levels(data1[,i])) )
	}
}
#
if (l3==1) {
	data2$con_rep(1,n) }
else if (l3==2) {
	data2[,2:p2]<-w3
	names(data2)[2:p2]<-names(w3)
	for (i in 2:p2) {
		if ( is.factor(data2[,i])==T ){
			contrasts(data2[,i])<-contr.sum(length(levels(data2[,i])))
		}
	}	
}
#
# Set formulas
if (constant[1]==1) { 
	data1$dcon<- c( rep(1,n),rep(0,n) )
	formula1<-'y0~.-v1' }
else { formula1<-'y0~.-1-v1' }
#
if      (l3==0) { formula0 <- 's~1'}
else if (l3==1) { formula0 <- 's~1'}
else            { 
	if (constant[2]==1) {formula0 <- 's~.-v1'}
	else {formula0 <- 's~.-1-v1'}
}
#
# Initial values for lambda
data1$v1<-1-c(vi,vi);
lambda<-glm(formula1,family=poisson, data=data1, weights=v1)$fitted
#
lambda1<-lambda[1:n]
lambda2<-lambda[(n+1):(2*n)]
#
difllike<-100.0
loglike0<-1000.0
i<-0
ii<-0

if (l3==0) {
	#
	# fit double poisson diagonal inflated model
	loglike<-rep(0,nmax)
	lambda3_1:n*0
	while ( (difllike>pres) & (i <= nmax) ) {
	i<-i+1
	#####   E step  ######
	for (j in 1:n) {
		if (zero[j]) {
			s[j]<-0;		
#			calculation of log-likelihood			
			if (x[j]==y[j]) {
				density.di<-di.f( 0.0, theta )
				like[j]<- log( (1-prob)*exp(-lambda1[j]-lambda2[j])+prob*density.di );
				vi[j]<-prob*density.di*exp(-like[j]) }
			else{
			        like[j]<-log(1-prob)+log(dpois(x[j],lambda1[j])) 																					 +log(dpois(y[j],lambda2[j]));
				vi[j]<-0.0 ;}
		}			
		else {
			if (x[j]==y[j]) {
					density.di_di.f( x[j],theta );
					like[j]<-log( (1-prob)*dpois( x[j],lambda1[j] )*dpois( y[j],lambda2[j] ) + prob*density.di );
					vi[j]  <- prob*density.di*exp( -like[j] ) }
			else	{
					vi[j]<-0.0;
					like[j]<-log(1-prob)+log( dpois(x[j],lambda1[j])*dpois(y[j],lambda2[j]) )}
		}
	}
#### end of E-step #########
	x1<-x;
	x2<-y;
	loglike[i]<-sum( like ) ; 
	difllike<-abs( (loglike0-loglike[i])/loglike0 )
	loglike0<-loglike[i]
	#
	#
########### M-step ############
	# estimate mixing proportion
	prob<-sum(vi)/n
	#
	# maximization of each theta parameter
	if ( distribution == 1 ) {
		# calculation of theta_j, j=1,...,jmax ; theta_0=1-sum(theta)
		for (ii in 1:jmax) {
			temp<-as.numeric(( (x==ii) & (y==ii) ));
			theta[ii]<-sum(temp*vi)/sum(vi) 
			}
			}
	else if (distribution==2){		
			# calculation of theta for poisson diagonal inflation
			theta<- sum(vi*x)/sum(vi) }
	else if (distribution==3){		
#	else {		
			# calculation of theta for geometric diagonal inflation
			theta<- sum(vi)/( sum(vi*x)+sum(vi) ) }
	#
	#	fit model on lambda1 & lambda2
	data1$v1<- 1-c(vi,vi);
	data1$y0<-c(x1,x2)
	m<-glm( formula1, family=poisson, data=data1, weights=v1 )

	p3<-length(m$coef)
	if (constant[1]==1) {
		beta<-c( m$coef[1]+m$coef[p3], m$coef[1:(p3-1)])
		names(beta)[1]<-'Intercept_1'
		names(beta)[2]<-'Intercept_2'	}
	else { beta<-m$coef }		
	lambda<-fitted(m)
	lambda1<-lambda[1:n]
	lambda2<-lambda[(n+1):(2*n)]
	#
	#####   end of M step  ######
	printvec<- c( i,beta,100.0*prob, theta,  loglike[i], difllike );	
	names(printvec)<-c( 'iter', names(beta),
                            'Mix.p(%)', paste( 'theta', 1:length(theta),sep='' ), 'loglike', 
			    'Rel.Dif.loglike')
	lengthprintvec<-length(printvec)
	print( c( round(printvec[1:(lengthprintvec-1)],2), printvec[lengthprintvec]) )
	}

#
#	calculation of BIC and AIC for double poisson model
	if ( (distribution==1)&(jmax==0) ){noparams<- m$rank +1}
	else                              {noparams<- m$rank + length( theta ) +1}
	AIC<- -2*loglike[i] + noparams * 2
	BIC<- -2*loglike[i] + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-0.000000000001
	y.mean<-y
	y.mean[y==0]<-0.000000000001
	AIC.sat <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#
	AICtotal<-c(AIC.sat, AIC); 
	BICtotal<-c(BIC.sat, BIC );
	names(AICtotal)<-c('Saturated', 'BP model')
	names(BICtotal)<-c('Saturated', 'BP model')
#
	if (l3==1) {lambda3<-lambda3[1]}
	result<-list(beta=beta, diagonal.distribution=dilabel, 
				p=prob, theta=theta, lambda1=m$fitted[1:n], 
				lambda2=m$fitted[(n+1):(2*n)],loglikelihood=loglike[1:i],
				parameters=noparams, AIC=AICtotal, BIC=BICtotal,iterations=i )
#
#
# end of diagonal inflated double poisson model
}
else {	
	loglike<-rep(0,nmax)
	while ( (difllike>pres) & (i <= nmax) ) {
	i<-i+1
	#####   E step  ######
	for (j in 1:n) {
		if (zero[j]) {
			s[j]<-0;		
#			calculation of log-likelihood			
			if (x[j]==y[j]) {
				density.di<-di.f( 0.0, theta )
				like[j]<- log( (1-prob)*exp(-lambda1[j]-lambda2[j]-lambda3[j])+prob*density.di );
				vi[j]<-prob*density.di*exp(-like[j]) }
			else{
			        like[j]<-log(1-prob)-lambda3[j] +log(dpois(x[j],lambda1[j])) 									+log(dpois(y[j],lambda2[j]));
				vi[j]<-0.0 ;}
		}			
		else {
			lbp1<-logbp(x[j]-1, y[j]-1, lambda=c(lambda1[j],lambda2[j],lambda3[j]) );
			lbp2<-logbp(x[j]  , y[j]  , lambda=c(lambda1[j],lambda2[j],lambda3[j]) );	
			s[j]<-exp( log(lambda3[j]) + lbp1 - lbp2 );
#			like[j]<-lbp2;
			if (x[j]==y[j]) {
					density.di_di.f( x[j],theta );
					like[j]<-log( (1-prob)*exp(lbp2) + prob*density.di );
					vi[j]  <- prob*density.di*exp( -like[j] ) }
			else	{
					vi[j]<-0.0;
					like[j]<-log(1-prob)+lbp2 }
		}
	}
#### end of E-step #########
	x1<-x-s;
	x2<-y-s;
	loglike[i]<-sum( like ) ; 
	difllike<-abs( (loglike0-loglike[i])/loglike0 )
	loglike0<-loglike[i]
	#
	#
########### M-step ############
	# estimate mixing proportion
	prob<-sum(vi)/n
	#
	# maximization of each theta parameter
	if ( distribution == 1 ) {
		# calculation of theta_j, j=1,...,jmax ; theta_0=1-sum(theta)
		for (ii in 1:jmax) {
			temp<-as.numeric(( (x==ii) & (y==ii) ));
			theta[ii]<-sum(temp*vi)/sum(vi) 
			}
			}
	else if (distribution==2){		
			# calculation of theta for poisson diagonal inflation
			theta<- sum(vi*x)/sum(vi) }
	else if (distribution==3){		
#	else {		
			# calculation of theta for geometric diagonal inflation
			theta<- sum(vi)/( sum(vi*x)+sum(vi) ) }
	# 	fit model on lambda3
	data2$v1<- 1-vi;
	data2$s<-s;
	m0<-glm( formula0, family=poisson, data=data2, weights=v1 )
	beta3<-m0$coef
	lambda3<-m0$fitted
	#
	#	fit model on lambda1 & lambda2
	data1$v1<- 1-c(vi,vi);
	data1$y0<-c(x1,x2)
	m<-glm( formula1, family=poisson, data=data1, weights=v1 )
	p3<-length(m$coef)
	if (constant[1]==1) {
		beta<-c( m$coef[1]+m$coef[p3], m$coef[1:(p3-1)])
		names(beta)[1]<-'Intercept_1'
		names(beta)[2]<-'Intercept_2'	}
	else {beta<-m$coef		}
	lambda<-fitted(m)
	lambda1<-lambda[1:n]
	lambda2<-lambda[(n+1):(2*n)]
	#
	#####   end of M step  ######
	printvec<- c( i,beta,beta3,100.0*prob, theta,  loglike[i], difllike );	
	names(printvec)<-c( 'iter', names(beta),paste('l3_',names(beta3),sep=''),
                            'Mix.p(%)', paste( 'theta', 1:length(theta),sep='' ), 'loglike', 
			    'Rel.Dif.loglike')
	lengthprintvec<-length(printvec)
	print( c( round(printvec[1:(lengthprintvec-1)],2), printvec[lengthprintvec]) )
	}
#
#	calculation of BIC and AIC for bivpoisson model
	if ( (distribution==1)&(jmax==0) ){noparams<- m$rank + m0$rank + 1}
	else                              {noparams<- m$rank + m0$rank + length( theta ) +1}
	AIC<- -2*loglike[i] + noparams * 2
	BIC<- -2*loglike[i] + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-0.000000000001
	y.mean<-y
	y.mean[y==0]<-0.000000000001
	AIC.sat <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#
	AICtotal<-c(AIC.sat, AIC); 
	BICtotal<-c(BIC.sat, BIC );
	names(AICtotal)<-c('Saturated', 'BP model')
	names(BICtotal)<-c('Saturated', 'BP model')
#
	if (l3==1) {lambda3<-lambda3[1]}
	result<-list(beta=beta, beta3=beta3, , diagonal.distribution=dilabel, 
				p=prob, theta=theta, lambda3=lambda3, lambda1=m$fitted[1:n], 
				lambda2=m$fitted[(n+1):(2*n)],loglikelihood=loglike[1:i],
				parameters=noparams, AIC=AICtotal, BIC=BICtotal,iterations=i )
#
} # end of elseif
#
result
#
#
}






